###############################################################################
# (c) Copyright 2020 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "COPYING".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
import os
import shutil
import subprocess
from os.path import dirname, join
from pathlib import Path

import pytest
import requests

EXAMPLE_REPO_ROOT = join(dirname(__file__), "data", "data-pkg-repo")
B02DKPi_versions = [
    "v0r0p2180615",
    "v0r0p2297458",
    "v0r0p2358939",
    "v0r0p2437974",
    "v0r0p2518507",
    "v0r0p2552694",
    "v0r0p2752886",
    "v0r0p2794507",
    "v0r0p2847846",
    "v0r0p2970193",
    "v0r0p3234056",
    "v0r0p3675801",
]


def pytest_addoption(parser):
    parser.addoption(
        "--runslow", action="store_true", default=False, help="run slow tests"
    )


def pytest_configure(config):
    config.addinivalue_line("markers", "slow: mark test as slow to run")


def pytest_collection_modifyitems(config, items):
    if config.getoption("--runslow"):
        # --runslow given in cli: do not skip slow tests
        return
    skip_slow = pytest.mark.skip(reason="need --runslow option to run")
    for item in items:
        if "slow" in item.keywords:
            item.add_marker(skip_slow)


@pytest.fixture
def empty_data_pkg_repo(tmpdir):
    cwd = os.getcwd()
    os.chdir(tmpdir)
    yield tmpdir
    os.chdir(cwd)


@pytest.fixture
def data_pkg_repo(empty_data_pkg_repo):
    shutil.copytree(EXAMPLE_REPO_ROOT, empty_data_pkg_repo, dirs_exist_ok=True)
    yield empty_data_pkg_repo


@pytest.fixture
def with_proxy(mocker):
    from subprocess import check_call as original_check_call

    def my_check_call(args, *more_args, **kwargs):
        if args == ["lb-dirac", "dirac-proxy-info", "--checkvalid"]:
            return
        else:
            return original_check_call(args, *more_args, **kwargs)

    mocker.patch("subprocess.check_call", new=my_check_call)


@pytest.fixture
def without_proxy(mocker):
    mocker.patch(
        "subprocess.check_call", side_effect=subprocess.CalledProcessError(1, "mocked")
    )


@pytest.fixture
def with_data(mocker, monkeypatch):
    from subprocess import run as original_run

    monkeypatch.setenv("LBRUN_BINDS", Path(__file__).parent)

    data_fn1 = Path(__file__).parent / "data" / "00070793_00000368_7.AllStreams.dst"
    if not data_fn1.is_file():
        url = (
            f"http://s3.cern.ch/lhcb-analysis-productions-dev/test-data/{data_fn1.name}"
        )
        with requests.get(url, stream=True) as r, open(data_fn1, "wb") as f:
            shutil.copyfileobj(r.raw, f)

    data_fn2 = Path(__file__).parent / "data" / "00070767_00000368_7.AllStreams.dst"
    if not data_fn2.is_file():
        url = (
            f"http://s3.cern.ch/lhcb-analysis-productions-dev/test-data/{data_fn2.name}"
        )
        with requests.get(url, stream=True) as r, open(data_fn2, "wb") as f:
            shutil.copyfileobj(r.raw, f)

    pool_catalog_xml = (
        '<?xml version="1.0" encoding="UTF-8" standalone="no" ?>\n'
        "<!-- Edited By PoolXMLCatalog.py -->\n"
        '<!DOCTYPE POOLFILECATALOG SYSTEM "InMemory">\n'
        "<POOLFILECATALOG>\n"
        '<File ID="980B119E-9609-E811-A58E-02163E016452">\n'
        "<physical>\n"
        f'    <pfn filetype="ROOT_All" name="{data_fn1}" se="CI-DST"/>\n'
        "</physical>\n"
        "<logical>\n"
        '    <lfn name="/lhcb/MC/2016/ALLSTREAMS.DST/00070793/0000/00070793_00000368_7.AllStreams.dst"/>\n'
        "</logical>\n"
        "</File>\n"
        '<File ID="043C600A-8F09-E811-B65B-FA163E9FB5C5">\n'
        "<physical>\n"
        f'    <pfn filetype="ROOT_All" name="{data_fn2}" se="CI-DST"/>\n'
        "</physical>\n"
        "<logical>\n"
        '    <lfn name="/lhcb/MC/2016/ALLSTREAMS.DST/00070767/0000/00070767_00000368_7.AllStreams.dst"/>\n'
        "</logical>\n"
        "</File>\n"
        "</POOLFILECATALOG>\n"
    )

    def my_run(args, *more_args, **kwargs):
        if args[:2] == ["lb-dirac", "dirac-bookkeeping-genXMLCatalog"]:
            with open(join(kwargs["cwd"], "pool_xml_catalog.xml"), "wt") as fp:
                fp.write(pool_catalog_xml)
            return subprocess.CompletedProcess([], 0, "")
        else:
            return original_run(args, *more_args, **kwargs)

    mocker.patch("subprocess.run", new=my_run)

    yield pool_catalog_xml


@pytest.fixture(scope="module")
def vcr_config():
    return {
        "filter_headers": ["authorization", "set-cookie"],
        "ignore_hosts": ["auth.cern.ch", "s3.cern.ch"],
    }


def monkeypatch_productions_to_versions(
    working_group: str, production_name: str
) -> list:
    return B02DKPi_versions


@pytest.fixture(autouse=True)
def production_to_versions_patched(monkeypatch):
    import LbAPLocal.cli

    monkeypatch.setattr(
        LbAPLocal.cli, "production_to_versions", monkeypatch_productions_to_versions
    )


@pytest.fixture
def fake_global_git_config(tmp_path, monkeypatch):
    """Create a fake global git config file."""
    git_config = tmp_path / "gitconfig"
    lines = [
        "[user]",
        "    name = Test User",
        "    email = test.user@domain.invalid",
    ]
    git_config.write_text("\n".join(lines))
    monkeypatch.setenv("GIT_CONFIG_GLOBAL", str(git_config))



###############################################################################
# (c) Copyright 2020-2024 CERN for the benefit of the LHCb Collaboration      #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "COPYING".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
import importlib
import io
import json
import os
import shlex
import shutil
import subprocess
import tempfile
import zipfile
from os.path import join

import click
import requests
from apd.authentication import get_auth_headers
from LbAPCommon import parse_yaml, render_yaml, validate_yaml
from LbAPCommon.hacks import project_uses_cmt
from LbAPCommon.options_parsing import validate_options

from .utils import (
    LBAPI_HOST,
    check_production,
    check_status_code,
    create_output_dir,
    pool_xml_catalog,
    recursively_create_input,
    resolve_input_overrides,
)


def prepare_test(production_name, job_name, dependent_input=None, n_events=None):
    """Run a local test job for a specific production job"""
    # Check if production exists
    check_production(production_name)

    # Check if job actually exists in production
    with open(os.path.join(production_name, "info.yaml"), "rt") as fp:
        raw_yaml = fp.read()
    prod_data = parse_yaml(render_yaml(raw_yaml))
    validate_yaml(prod_data, ".", production_name)
    try:
        job_data = prod_data[job_name]
    except KeyError:
        raise click.ClickException(
            f"Job {job_name} is not found for production {production_name}!"
        )

    application_name, application_version = job_data["application"].rsplit("/", 1)
    application = {
        "name": application_name,
        "version": application_version,
        "data_pkgs": ["AnalysisProductions v999999999999"],
    }
    if "@" in application_version:
        app_version, binary_tag = application_version.split("@", 1)
        application["version"] = app_version
        application["binary_tag"] = binary_tag

    require_cmt = project_uses_cmt(application["name"], application["version"])

    prodrun_conf = {
        "spec_version": 1,
        "application": application,
        "input": {"xml_summary_file": "summaryDaVinci_00012345_00006789_1.xml"},
        "output": {"prefix": "00012345_00006789_1"},
    }

    if isinstance(job_data["options"], dict):
        prodrun_conf["options"] = job_data["options"].copy()
        if "entrypoint" not in job_data["options"]:
            prodrun_conf["options"].setdefault("format", "WGProd")
    else:
        prodrun_conf["options"] = {"format": "WGProd", "files": job_data["options"]}

    if n_events is not None:
        prodrun_conf["input"]["n_of_events"] = n_events

    extra_files = []
    override_input, output_filetype = resolve_input_overrides(prod_data, job_name)
    if "job_name" in job_data["input"] and dependent_input is None:
        dependent_input, extra_files = recursively_create_input(
            production_name,
            job_data["input"]["job_name"],
            job_data["input"].get("filetype"),
            prod_data,
            n_events,
        )

    # only create output directories if this is the job that's about to be run
    # i.e. there are no dependent jobs that need to be run first
    dynamic_dir, out_dir = create_output_dir(production_name, require_cmt)
    for path in extra_files:
        shutil.copyfile(path, f"{out_dir}/{path.name}")

    if job_data["automatically_configure"]:
        if "files" not in prodrun_conf["options"]:
            raise NotImplementedError(
                "automatically_configure is not yet supported for lbexec based productions"
            )
        params = {}
        if output_filetype:
            params["override-output-filetype"] = output_filetype
        response = requests.post(
            f"{LBAPI_HOST}/pipelines/autoconf-options/",
            **get_auth_headers(),
            json={**job_data, **({"input": override_input} if override_input else {})},
            params=params,
        )
        check_status_code(response)

        config_path = dynamic_dir / production_name / f"{job_name}_autoconf.py"
        config_path.parent.mkdir(parents=True)
        config_path.write_text(response.text)
        prodrun_conf["options"]["files"].insert(
            0, join("$ANALYSIS_PRODUCTIONS_DYNAMIC", production_name, config_path.name)
        )

    prodrun_config_path = out_dir / "prodConf_DaVinci_00012345_00006789_1.json"

    prodrun_conf["output"]["types"] = job_data["output"]

    # force to use the file chosen by the user
    if "bk_query" in job_data["input"] or "transform_ids" in job_data["input"]:
        if dependent_input is not None:
            prodrun_conf["input"]["files"] = [dependent_input]
            if "LFN:" in dependent_input:
                (out_dir / "pool_xml_catalog.xml").write_text(
                    pool_xml_catalog([dependent_input])
                )
                prodrun_conf["input"]["xml_file_catalog"] = "pool_xml_catalog.xml"
        else:
            response = requests.post(
                f"{LBAPI_HOST}/pipelines/lfns/",
                **get_auth_headers(),
                json=job_data["input"],
            )
            lfns = check_status_code(response).json()

            prodrun_conf["input"]["files"] = [f"LFN:{lfn}" for lfn in lfns]
            (out_dir / "pool_xml_catalog.xml").write_text(pool_xml_catalog(lfns))
            prodrun_conf["input"]["xml_file_catalog"] = "pool_xml_catalog.xml"

    elif "job_name" in job_data["input"]:
        prodrun_conf["input"]["files"] = [dependent_input]

    prodrun_config_path.write_text(json.dumps(prodrun_conf))
    return out_dir, prodrun_config_path.name


def prepare_reproduce(pipeline_id, production_name, job_name, test_id):
    click.secho(
        f"Reproducing test for test {pipeline_id} {production_name} {job_name}",
        fg="green",
    )

    prod_url = f"{LBAPI_HOST}/pipelines/{pipeline_id}/{production_name}/"
    response = requests.get(prod_url, **get_auth_headers())
    prod_info = check_status_code(response).json()

    job_url = f"{prod_url}jobs/{job_name}/"
    if test_id != "latest":
        job_url += f"tests/{test_id}"
    response = requests.get(job_url, **get_auth_headers())
    job_info = check_status_code(response).json()
    if test_id == "latest":
        job_url += f"tests/{job_info['test']['attempt']}"

    tmp_dir = tempfile.mkdtemp()

    click.secho(f"Cloning {prod_info['repo_url']}", fg="green")
    subprocess.check_call(["git", "clone", prod_info["repo_url"], tmp_dir])

    click.secho(f"Running test in {tmp_dir}", fg="green")
    os.chdir(tmp_dir)

    click.secho(f"Fetching and checking out {prod_info['commit']}", fg="green")
    subprocess.check_call(["git", "fetch", "origin", prod_info["commit"]])
    subprocess.check_call(["git", "checkout", prod_info["commit"]])

    check_production(production_name)

    require_cmt = project_uses_cmt(
        job_info["application_name"], job_info["application_version"]
    )
    dynamic_dir, out_dir = create_output_dir(production_name, require_cmt)

    # Download the jobs output sandbox
    response = requests.get(f"{job_url}/zip", **get_auth_headers())
    zip_data = io.BytesIO(check_status_code(response).content)
    with zipfile.ZipFile(zip_data) as zf:
        zf.extractall(out_dir)
    prodrun_config_path = max(
        out_dir.glob("prodConf*.json"), key=lambda x: x.name.split("_")[-1]
    )

    # Write automatic configuration if requested
    if job_info["automatically_configure"]:
        filename = dynamic_dir / job_info["dynamic_options_path"]
        # Prevent directory traversal
        if str(filename.relative_to(dynamic_dir)).startswith("."):
            raise NotImplementedError(dynamic_dir, filename)
        filename.parent.mkdir(parents=True)
        filename.write_text(job_info["autoconf_options"])

    # Download the job input
    input_paths = [x["path"] for x in job_info["test"]["input_files"]]
    if not input_paths:
        raise NotImplementedError("Failed to find any input files")
    if all("/" in x for x in input_paths):
        # It's an LFN so write the pool_xml_catalog.xml
        (out_dir / "pool_xml_catalog.xml").write_text(pool_xml_catalog(input_paths))
    elif all("/" not in x for x in input_paths):
        if len(input_paths) != 1:
            raise NotImplementedError(input_paths)
        filename = out_dir / input_paths[0]
        # It's from another job so download from S3
        response = requests.get(
            f"{prod_url}jobs/{job_info['input']['job_name']}/tests/"
            f"{job_info['test']['attempt']}/files/{filename.name}/url",
            **get_auth_headers(),
        )
        url = check_status_code(response).json()["url"]

        click.secho(f"Downloading {filename.name}", fg="green")
        with requests.get(url, stream=True) as resp:
            check_status_code(resp)
            with filename.open("wb") as fp:
                shutil.copyfileobj(resp.raw, fp)

        prodrun_conf = json.loads(prodrun_config_path.read_text())
        prodrun_conf["input"]["files"] = [filename.name]
        prodrun_config_path.write_text(json.dumps(prodrun_conf))
    else:
        raise NotImplementedError(input_paths)

    return out_dir, prodrun_config_path.name


def enter_debugging(out_dir, prodrun_config_path, dependent_job=None):
    # This means the requested job depends on another job and has not been told
    # the location of that jobs output
    if dependent_job is not None:
        raise NotImplementedError(
            "Automatic input creation for job-dependent jobs is not implemented in debug mode \n"
            f"The requested job depends on {dependent_job['job_name']}, please provide the location "
            f"of the output file of a local test of {dependent_job['job_name']} by appending "
            '" -i <file_location>" to the debug command'
        )
    cmd = ["lb-prod-run", prodrun_config_path, "--interactive"]
    click.secho(f"Running: lb-prod-run {shlex.join(cmd)}")
    os.chdir(out_dir)
    os.execlp(cmd[0], *cmd)


def do_options_parsing(env_cmd, out_dir, pkl_file, root_file, job_name, prod_data):
    json_file = join(out_dir, "output.json")

    with importlib.resources.path(
        "LbAPCommon.options_parsing", "gaudi_pickle_to_json.py"
    ) as script_path:
        subprocess.run(
            env_cmd
            + [
                "python",
                script_path,
                "--pkl",
                pkl_file,
                "--output",
                json_file,
                "--debug",
            ],
            text=True,
        )
    return validate_options(json_file, root_file, job_name, prod_data)



###############################################################################
# (c) Copyright 2020 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "COPYING".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
import os
from os.path import isfile, join

import pytest
from click.testing import CliRunner

import LbAPLocal
import LbAPLocal.testing
from LbAPLocal.cli import main


@pytest.mark.slow
@pytest.mark.vcr
@pytest.mark.parametrize(
    "name,n_events",
    [
        ("2016_MagDown_PromptMC_D02KK", 2444),
        ("2016_MagUp_PromptMC_D02KK", 2393),
        ("2016_MagDown_PromptMC_D02KK_noautoconf", 2444),
        ("2016_MagUp_PromptMC_D02KK_noautoconf", 2393),
    ],
)
def test_run_test(
    with_proxy, data_pkg_repo, with_data, name, n_events, monkeypatch, record_mode
):
    if record_mode == "none":
        monkeypatch.setattr(
            LbAPLocal.cli,
            "get_auth_headers",
            lambda: dict(headers={"Authorization": "Bearer "}),
        )
        monkeypatch.setattr(
            LbAPLocal.testing,
            "get_auth_headers",
            lambda: dict(headers={"Authorization": "Bearer "}),
        )

    runner = CliRunner()
    result = runner.invoke(main, ["test", "executabletests", name, "--no-validate"])
    assert result.exit_code == 0, result.stdout

    assert "Application Manager Finalized successfully" in result.stdout
    assert "Application Manager Terminated successfully" in result.stdout
    assert f"{n_events} events processed" in result.stdout
    assert "Summary of log messages" in result.stdout
    assert "General explanations" in result.stdout
    assert "Histograms are not being saved" in result.stdout

    assert "Output can be found in" in result.stdout
    output_dir = result.stdout.split("Output can be found in")[-1].strip()

    output_fn = join(output_dir, "00012345_00006789_1.D02KK.ROOT")
    assert isfile(output_fn)
    assert 100 * 1024 <= os.stat(output_fn).st_size <= 1024**2

    stdout_fn = join(output_dir, "stdout.log")
    assert isfile(stdout_fn)
    with open(stdout_fn, "rt") as fp:
        stdout = fp.read()
    assert "Application Manager Finalized successfully" in stdout
    assert "Application Manager Terminated successfully" in stdout
    assert f"{n_events} events processed" in stdout

    stderr_fn = join(output_dir, "stderr.log")
    assert isfile(stderr_fn)


# TODO: Re-add this
# @pytest.mark.slow
# @pytest.mark.vcr
# def test_reproduce(with_proxy, with_data, mocker):
#     from LbAPLocal.testing import prepare_reproduce

#     prepare_reproduce("3558905", "cbtesting", "2016_MagUp_PromptMC_D02KK", "0")

#     execlp_mock = mocker.patch("os.execlp", side_effect=None)

#     runner = CliRunner()
#     result = runner.invoke(
#         main, ["reproduce", "3558905", "cbtesting", "2016_MagUp_PromptMC_D02KK", "0"]
#     )
#     assert result.exit_code == 0, result.stdout
#     assert "Cloning " in result.stdout
#     assert "Downloading pool_xml_catalog.xml" in result.stdout
#     assert "Downloading prodConf_DaVinci_00012345_00006789_1.py" in result.stdout
#     assert "Starting lb-run with: lb-run" in result.stdout

#     assert execlp_mock.call_count == 1
#     cmd, *execlp_args = execlp_mock.call_args[0]
#     assert cmd == execlp_args[0]
#     assert "--use=ProdConf" in execlp_args
#     assert "DaVinci/v39r1p6" in execlp_args

#     assert isfile(execlp_args[execlp_args.index("--rcfile") + 1])
#     with subprocess.Popen(
#         execlp_args + ["-i"],
#         stdout=subprocess.PIPE,
#         stdin=subprocess.PIPE,
#         stderr=subprocess.PIPE,
#     ) as p:
#         stdout_data, stderr_data = p.communicate(input=b"which gaudirun.py")
#     stdout_data = list(filter(None, stdout_data.decode().strip().split("\n")))
#     assert "Welcome to analysis productions debug mode:" in stdout_data
#     assert (
#         stdout_data[-1]
#         == "/cvmfs/lhcb.cern.ch/lib/lhcb/GAUDI/GAUDI_v27r0/InstallArea/x86_64-slc6-gcc49-opt/scripts/gaudirun.py"
#     )
#     assert not isfile(execlp_args[execlp_args.index("--rcfile") + 1])

#     gaudirun_cmd = stdout_data[-2]
#     assert gaudirun_cmd.startswith("gaudirun.py ")
